#!python
# cython: infer_types=True


# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.

cimport cython
cimport numpy as np
import numpy as np
import unittest
from cpython cimport bool
from libc.stdlib cimport rand, RAND_MAX
IF UNAME_SYSNAME == "Windows":
    from libc.math cimport HUGE_VAL as inf
ELSE:
    from libc.math cimport INFINITY as inf
    from libc.math cimport fmin
    from libc.math cimport erf
    from libc.math cimport tgamma
from libc.math cimport sqrt
from libc.math cimport exp
from libc.math cimport pow
from libc.math cimport fabs
from libc.math cimport fmod
from libc.math cimport cos
from libc.math cimport acos
from libc.math cimport sin
from libc.math cimport M_PI as pi
from libc.math cimport M_SQRT2 as sqrt2

from MTfit.convert.cmoment_tensor_conversion cimport cTape_MT6

ctypedef double DTYPE_t
cdef DTYPE_t PI2=2*pi  


ctypedef DTYPE_t (*jump_params_ptr)(DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t)
ctypedef DTYPE_t (*prior_ratio_ptr)(DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t) nogil
ctypedef DTYPE_t (*transition_ratio_ptr)(DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t,DTYPE_t) nogil

#Windows specific functions - in libc.math otherwise
IF UNAME_SYSNAME == "Windows":    
    cdef DTYPE_t ND=1902.9928503773808*1.10452194071529090000 #Calculated from python and MATLAB
    
    #Approx for erf function
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef inline double erf(double x) nogil:
        # constants
        cdef double a1 =  0.254829592
        cdef double a2 = -0.284496736
        cdef double a3 =  1.421413741
        cdef double a4 = -1.453152027
        cdef double a5 =  1.061405429
        cdef double p  =  0.3275911
        # Save the sign of x
        cdef int sign = 1
        if x < 0:
            sign = -1
        x = fabs(x)
        # A&S formula 7.1.26
        cdef double t = 1.0/(1.0 + p*x)
        cdef double y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*exp(-x*x)
        return sign*y
    
    #no fmin so acceptance different to *nix
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef DTYPE_t acceptance(transition_ratio_ptr transition_ratio_fn,prior_ratio_ptr prior_ratio_fn,jump_params_ptr jump_params_fn,DTYPE_t gamma,DTYPE_t delta,DTYPE_t h,DTYPE_t sigma,DTYPE_t g0,DTYPE_t g_s,DTYPE_t d0,DTYPE_t d_s,DTYPE_t h0,DTYPE_t h_s,DTYPE_t s0,DTYPE_t s_s,DTYPE_t ln_p,DTYPE_t ln_p0,int jump,DTYPE_t qg,DTYPE_t qd,DTYPE_t sg,DTYPE_t sd,DTYPE_t proposal_normalisation,DTYPE_t p_dc) :
        cdef DTYPE_t acc
        if jump>0:
            if gamma==0.0 and delta==0.0:
                #MT to DC
                model_prior_ratio=p_dc/(1-p_dc)
                acc=exp(ln_p-ln_p0)*prior_ratio_fn(gamma,delta,g0,d0)*jump_params_fn(qg,qd,sg,sd, proposal_normalisation)*model_prior_ratio
            elif g0==0.0 and d0==0.0 :
                #DC to MT
                model_prior_ratio=(1-p_dc)/p_dc
                acc=exp(ln_p-ln_p0)*prior_ratio_fn(gamma,delta,g0,d0)/jump_params_fn(qg,qd,sg,sd,proposal_normalisation)*model_prior_ratio            
        else:   
            acc=exp(ln_p-ln_p0)*transition_ratio_fn(gamma,delta,h,sigma,g0,g_s,d0,d_s,h0,h_s,s0,s_s)*prior_ratio_fn(gamma,delta,g0,d0)
        if acc<1:
            return acc
        return 1
ELSE:
    cdef DTYPE_t ND=tgamma(11.490)/(tgamma(5.745)*tgamma(5.745))*1.10452194071529090000

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef DTYPE_t acceptance(transition_ratio_ptr transition_ratio_fn,prior_ratio_ptr prior_ratio_fn,jump_params_ptr jump_params_fn,DTYPE_t gamma,DTYPE_t delta,DTYPE_t h,DTYPE_t sigma,DTYPE_t g0,DTYPE_t g_s,DTYPE_t d0,DTYPE_t d_s,DTYPE_t h0,DTYPE_t h_s,DTYPE_t s0,DTYPE_t s_s,DTYPE_t ln_p,DTYPE_t ln_p0,int jump,DTYPE_t qg,DTYPE_t qd,DTYPE_t sg,DTYPE_t sd,DTYPE_t proposal_normalisation,DTYPE_t p_dc) :
        if jump>0:
            if gamma==0.0 and delta==0.0:
                model_prior_ratio=p_dc/(1-p_dc)
                return fmin(1,exp(ln_p-ln_p0)*prior_ratio_fn(gamma,delta,g0,d0)*jump_params_fn(qg,qd,sg,sd, proposal_normalisation)*model_prior_ratio)
            elif g0==0.0 and d0==0.0 :
                model_prior_ratio=(1-p_dc)/p_dc
                return fmin(1,exp(ln_p-ln_p0)*prior_ratio_fn(gamma,delta,g0,d0)/jump_params_fn(qg,qd,sg,sd,proposal_normalisation)*model_prior_ratio)
        return fmin(1,exp(ln_p-ln_p0)*transition_ratio_fn(gamma,delta,h,sigma,g0,g_s,d0,d_s,h0,h_s,s0,s_s)*prior_ratio_fn(gamma,delta,g0,d0))

#Gaussian PDF
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t gaussian_pdf(DTYPE_t x,DTYPE_t mu,DTYPE_t s) nogil:
    return (1/(sqrt2*sqrt(pi)*s))*exp(-(x-mu)*(x-mu)/(2*s*s))

#Gaussian CDF
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t gaussian_cdf(DTYPE_t x,DTYPE_t mu,DTYPE_t s) nogil:
    return 0.5*(1+erf((x-mu)/(s*sqrt2)))

#Inline functions for Gaussian transition ratios - normalising for truncated gaussian PDFs
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t gaussian_transition_mt(DTYPE_t gamma,DTYPE_t delta,DTYPE_t h,DTYPE_t sigma,DTYPE_t g0,DTYPE_t g_s,DTYPE_t d0,DTYPE_t d_s,DTYPE_t h0,DTYPE_t h_s,DTYPE_t s0,DTYPE_t s_s) nogil:
    N=(gaussian_cdf(pi/6,g0,g_s)-gaussian_cdf(-pi/6,g0,g_s))*(gaussian_cdf(pi/2,d0,d_s)-gaussian_cdf(-pi/2,d0,d_s))
    return gaussian_transition_dc(h,sigma,h0,h_s,s0,s_s)/N

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t gaussian_transition_dc(DTYPE_t h,DTYPE_t sigma,DTYPE_t h0,DTYPE_t h_s,DTYPE_t s0,DTYPE_t s_s) nogil:
    N=(gaussian_cdf(1,h0,h_s)-gaussian_cdf(0,h0,h_s))*(gaussian_cdf(pi/2,s0,s_s)-gaussian_cdf(-pi/2,s0,s_s))
    return 1/N

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t gaussian_transition_ratio(DTYPE_t gamma,DTYPE_t delta,DTYPE_t h,DTYPE_t sigma,DTYPE_t g0,DTYPE_t g_s,DTYPE_t d0,DTYPE_t d_s,DTYPE_t h0,DTYPE_t h_s,DTYPE_t s0,DTYPE_t s_s) nogil:
    if gamma==0.0 and delta==0.0 and g0==0.0 and d0==0.0:
        return gaussian_transition_dc(h0,s0,h,h_s,sigma,s_s)/gaussian_transition_dc(h,sigma,h0,h_s,s0,s_s)
    return gaussian_transition_mt(g0,d0,h0,s0,gamma,g_s,delta,d_s,h,h_s,sigma,s_s)/(gaussian_transition_mt(gamma,delta,h,sigma,g0,g_s,d0,d_s,h0,h_s,s0,s_s))

#Prior distributions - distribution for delta from numerical tests.
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t uniform_delta_dist(DTYPE_t delta) nogil:
    #based on beta dist
    b=5.745
    d=(delta+pi/2)/pi
    return ND*pow(d*(1-d),b-1)/pi

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
#Prior ratios, including DC/MT jumps for relative inversion
cdef inline DTYPE_t uniform_prior_ratio(DTYPE_t gamma,DTYPE_t delta,DTYPE_t g0,DTYPE_t d0) nogil:
    if gamma==0.0 and delta==0.0 and g0==0.0 and d0==0.0:
        return 1.0
    elif gamma==0.0 and delta==0.0:
        return 1/(1.5*cos(3*g0)*uniform_delta_dist(d0))
    elif g0==0.0 and d0==0.0:
        return 1.5*cos(3*gamma)*uniform_delta_dist(delta)
    return cos(3*gamma)*uniform_delta_dist(delta)/(cos(3*g0)*uniform_delta_dist(d0))

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t flat_prior_ratio(DTYPE_t gamma,DTYPE_t delta,DTYPE_t g0,DTYPE_t d0) nogil:
    return 1.0

# Functions for obtaining new Tape parameter samples etc.
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef DTYPE_t new_gamma(DTYPE_t g0, DTYPE_t g_s):
    cdef DTYPE_t gamma
    gamma=g_s*np.random.randn(1)+g0
    while fabs(gamma)>pi/6:
        gamma=g_s*np.random.randn(1)+g0
    return gamma

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef DTYPE_t new_delta(DTYPE_t d0, DTYPE_t d_s):
    cdef DTYPE_t delta
    delta=d_s*np.random.randn(1)+d0
    while fabs(delta)>pi/2:
        delta=d_s*np.random.randn(1)+d0
    return delta

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef DTYPE_t new_kappa(DTYPE_t k0, DTYPE_t k_s):
    cdef DTYPE_t kn=k_s*np.random.randn(1)+k0
    if fabs(kn)<PI2:
        kn=fmod(kn,PI2)
    if kn<0.0:
        kn+=PI2
    return kn

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef DTYPE_t new_h(DTYPE_t h0, DTYPE_t h_s):
    cdef DTYPE_t h
    h=h_s*np.random.randn(1)+h0
    while h>1 or h<0:
        h=h_s*np.random.randn(1)+h0
    return h

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef DTYPE_t new_sigma(DTYPE_t s0, DTYPE_t s_s):
    cdef DTYPE_t sigma
    sigma=s_s*np.random.randn(1)+s0
    while fabs(sigma)>pi/2:
        sigma=s_s*np.random.randn(1)+s0
    return sigma

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def convert_sample(DTYPE_t gamma,DTYPE_t  delta,DTYPE_t  kappa,DTYPE_t  h,DTYPE_t  sigma):
    cdef DTYPE_t[:,::1] M = np.empty((6,1))
    cdef DTYPE_t[::1] m = np.empty((6))
    cTape_MT6(&m[0],gamma,delta,kappa,h,sigma)
    M[0,0]=m[0]
    M[1,0]=m[1]
    M[2,0]=m[2]
    M[3,0]=m[3]
    M[4,0]=m[4]
    M[5,0]=m[5]
    return np.asarray(M)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t ranf() nogil:
    return <DTYPE_t>rand()/<DTYPE_t>RAND_MAX

#Acceptance check functions single and multiple events
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef int c_acceptance_check(transition_ratio_ptr transition_ratio_fn,prior_ratio_ptr prior_ratio_fn,jump_params_ptr jump_params_fn,DTYPE_t[::1] gamma,DTYPE_t[::1]  delta,DTYPE_t[::1]  h,DTYPE_t[::1]  sigma,DTYPE_t[::1] kappa,DTYPE_t g0,DTYPE_t g_s,DTYPE_t d0,DTYPE_t d_s,DTYPE_t h0,DTYPE_t h_s,DTYPE_t s0,DTYPE_t s_s,DTYPE_t[::1] ln_p,DTYPE_t ln_p0,DTYPE_t k0,DTYPE_t[::1] qg,DTYPE_t[::1] qd,DTYPE_t sg,DTYPE_t sd,DTYPE_t proposal_normalisation,DTYPE_t p_dc):
    cdef Py_ssize_t umax=ln_p.shape[0]
    cdef Py_ssize_t u
    cdef int jump=0
    cdef DTYPE_t acc=0
    for u in range(umax):
        jump=0
        if ((g0==0.0 and d0==0.0) or (gamma[u]==0.0 and delta[u]==0.0)) and  not gamma[u]==g0 and not delta[u]==d0 and h0==h[u] and s0==sigma[u] and k0==kappa[u]:
            jump=1
        if ln_p[u]==-inf:
            continue
        if acceptance(transition_ratio_fn,prior_ratio_fn,jump_params_fn,gamma[u], delta[u], h[u], sigma[u], g0, g_s, d0, d_s, h0, h_s, s0, s_s, ln_p[u], ln_p0,jump,qg[u],qd[u],sg,sd,proposal_normalisation,p_dc)>np.random.rand():
            return <int>u
    return -1

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef int c_me_acceptance_check(transition_ratio_ptr transition_ratio_fn,prior_ratio_ptr prior_ratio_fn,jump_params_ptr jump_params_fn,DTYPE_t[:,::1] gamma,DTYPE_t[:,::1]  delta,DTYPE_t[:,::1]  h,DTYPE_t[:,::1]  sigma,DTYPE_t[:,::1] kappa,DTYPE_t[::1] g0,DTYPE_t[::1] g_s,DTYPE_t[::1] d0,DTYPE_t[::1] d_s,DTYPE_t[::1] h0,DTYPE_t[::1] h_s,DTYPE_t[::1] s0,DTYPE_t[::1] s_s,DTYPE_t[::1] ln_p,DTYPE_t ln_p0,DTYPE_t[::1] k0,DTYPE_t[:,::1] qg,DTYPE_t[:,::1] qd,DTYPE_t[::1] sg,DTYPE_t[::1] sd,DTYPE_t[::1] proposal_normalisation,DTYPE_t[::1] p_dc):
    cdef Py_ssize_t umax=ln_p.shape[0]
    cdef Py_ssize_t u
    cdef Py_ssize_t nevents=gamma.shape[1]
    cdef Py_ssize_t v
    cdef int jump=0
    cdef DTYPE_t acc=1
    for u in range(umax):
        if ln_p[u]==-inf:
            continue
        for v in range(nevents):
            jump=0
            if ((g0[v]==0.0 and d0[v]==0.0) or (gamma[u,v]==0.0 and delta[u,v]==0.0)) and  not gamma[u,v]==g0[v] and not delta[u,v]==d0[v] and h0[v]==h[u,v] and s0[v]==sigma[u,v] and k0[v]==kappa[u,v]:#Check for jump
                jump=1
            acc*=acceptance(transition_ratio_fn,prior_ratio_fn,jump_params_fn,gamma[u,v], delta[u,v], h[u,v], sigma[u,v], g0[v], g_s[v], d0[v], d_s[v], h0[v], h_s[v], s0[v], s_s[v], 0,0,jump,qg[u,v],qd[u,v],sg[v],sd[v],proposal_normalisation[v],p_dc[v])
            if acc==0:
                break
        if v<nevents-1 or acc==0.0:#No acceptance, so no point in continuing
            continue
        acc*=exp(ln_p[u]-ln_p0)
        if acc>np.random.rand():
            return <int>u
    return -1

#RJ Jump paramters
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline tuple flat_jump_params(DTYPE_t a,DTYPE_t b,):
    return  (np.random.rand()*2-1)*pi/6,acos(np.random.rand()*2-1)-pi/2

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline tuple gaussian_jump_params(DTYPE_t sg,DTYPE_t sd):
    cdef DTYPE_t qg=-pi/2
    cdef DTYPE_t qd=-pi
    while fabs(qg)>pi/6:
        qg=sg*np.random.randn()
    while fabs(qd)>pi/2:
        qd=sd*np.random.randn()
    return  (qg,qd)

#RJ jump parameter priors
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t flat_jump_prob(DTYPE_t a,DTYPE_t b,DTYPE_t sg,DTYPE_t sd,DTYPE_t proposal_normalisation):
    #N.B. ILLOGICAL - DO NOT USE - WILL ALWAYS ACCEPT JUMP FROM DC to MT (gamma=0 delta=0), i.e. increasing complexity
    return  3/(PI2)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t gaussian_jump_prob(DTYPE_t qg,DTYPE_t qd,DTYPE_t sg,DTYPE_t sd,DTYPE_t proposal_normalisation):
    return  gaussian_pdf(qg,0,sg)*gaussian_pdf(qd,0,sd)/proposal_normalisation

#Python accessible functions for acceptance checks
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def acceptance_check(list x,dict x0,dict alpha, ln_p, DTYPE_t ln_p0, bool uniform_prior=True, bool gaussian_jump=True, DTYPE_t dc_prior=0.5):
    cdef Py_ssize_t umax=ln_p.shape[0]
    cdef Py_ssize_t u
    cdef DTYPE_t[::1] gamma=np.empty((umax))
    cdef DTYPE_t[::1] delta=np.empty((umax))
    cdef DTYPE_t[::1] h=np.empty((umax))
    cdef DTYPE_t[::1] sigma=np.empty((umax))
    cdef DTYPE_t[::1] kappa=np.empty((umax))   
    cdef DTYPE_t[::1] qg=np.empty((umax))
    cdef DTYPE_t[::1] qd=np.empty((umax))          
    cdef transition_ratio_ptr transition_ratio_fn=&gaussian_transition_ratio
    cdef prior_ratio_ptr prior_ratio_fn    
    cdef jump_params_ptr jump_params_fn
    cdef DTYPE_t g0=<DTYPE_t>x0['gamma']
    cdef DTYPE_t g_s=<DTYPE_t>alpha['gamma']
    cdef DTYPE_t d0=<DTYPE_t>x0['delta']
    cdef DTYPE_t d_s=<DTYPE_t>alpha['delta']
    cdef DTYPE_t h0=<DTYPE_t>x0['h']
    cdef DTYPE_t h_s=<DTYPE_t>alpha['h']
    cdef DTYPE_t s0=<DTYPE_t>x0['sigma']
    cdef DTYPE_t s_s=<DTYPE_t>alpha['sigma']
    cdef DTYPE_t k0=<DTYPE_t>x0['kappa']
    cdef DTYPE_t proposal_normalisation=1.0
    if 'proposal_normalisation' in alpha.keys():
        proposal_normalisation=<DTYPE_t>alpha['proposal_normalisation']
    cdef DTYPE_t sg=0.1
    cdef DTYPE_t sd=0.1
    cdef int index
    if gaussian_jump:
        jump_params_fn=&gaussian_jump_prob
    else:
        jump_params_fn=&flat_jump_prob
    if uniform_prior:
        prior_ratio_fn=&uniform_prior_ratio
    else:
        prior_ratio_fn=&flat_prior_ratio
    if 'gamma_dc' in alpha.keys():
        sg=<DTYPE_t>alpha['gamma_dc']
    if 'delta_dc' in alpha.keys():
        sd=<DTYPE_t>alpha['delta_dc']
    for u in range(umax):
        gamma[u]=x[u]['gamma']
        delta[u]=x[u]['delta']
        kappa[u]=x[u]['kappa']
        sigma[u]=x[u]['sigma']
        h[u]=x[u]['h']
        if 'g0' in x[u].keys():
            qg[u]=x[u]['g0']#g0 from mt sample ==> dc sample
        else:
            qg[u]=x[u]['gamma'];#no g0  ==>  mt sample
        if 'd0' in x[u].keys():
            qd[u]=x[u]['d0']
        else:
            qd[u]=x[u]['delta'];
    index=c_acceptance_check(transition_ratio_fn,prior_ratio_fn,jump_params_fn, gamma, delta, h, sigma,kappa, g0, g_s, d0, d_s, h0, h_s, s0, s_s,ln_p,ln_p0,k0,qg,qd,sg,sd,proposal_normalisation,dc_prior)
    if index>=0:
        return x[index],ln_p[index],index
    return {},False,umax

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def me_acceptance_check(list x,list x0,list alpha, DTYPE_t[::1] ln_p,DTYPE_t ln_p0,bool uniform_prior=True,bool gaussian_jump=True,DTYPE_t[::1] dc_prior=np.array([0.5])):
    cdef Py_ssize_t umax=ln_p.shape[0]
    cdef Py_ssize_t nevents=len(alpha)
    cdef Py_ssize_t u
    cdef Py_ssize_t v
    cdef DTYPE_t[:,::1] gamma=np.empty((umax,nevents))
    cdef DTYPE_t[:,::1] delta=np.empty((umax,nevents))
    cdef DTYPE_t[:,::1] h=np.empty((umax,nevents))
    cdef DTYPE_t[:,::1] sigma=np.empty((umax,nevents))
    cdef DTYPE_t[:,::1] kappa=np.empty((umax,nevents))   
    cdef DTYPE_t[:,::1] qg=np.empty((umax,nevents))
    cdef DTYPE_t[:,::1] qd=np.empty((umax,nevents))          
    cdef transition_ratio_ptr transition_ratio_fn=&gaussian_transition_ratio
    cdef prior_ratio_ptr prior_ratio_fn    
    cdef jump_params_ptr jump_params_fn
    cdef DTYPE_t [::1]g0=np.empty((nevents))
    cdef DTYPE_t [::1]g_s=np.empty((nevents))
    cdef DTYPE_t [::1]d0=np.empty((nevents))
    cdef DTYPE_t [::1]d_s=np.empty((nevents))
    cdef DTYPE_t [::1]h0=np.empty((nevents))
    cdef DTYPE_t [::1]h_s=np.empty((nevents))
    cdef DTYPE_t [::1]s0=np.empty((nevents))
    cdef DTYPE_t [::1]s_s=np.empty((nevents))
    cdef DTYPE_t [::1]k0=np.empty((nevents))
    cdef DTYPE_t [::1]sg=np.empty((nevents))
    cdef DTYPE_t [::1]sd=np.empty((nevents))
    cdef DTYPE_t [::1]proposal_normalisation=np.ones((nevents))
    cdef DTYPE_t [::1]prior_dc=np.ones((nevents))
    if dc_prior.shape[0]==0:
        prior_dc=np.array([0.5])
    cdef int index
    if gaussian_jump:
        jump_params_fn=&gaussian_jump_prob
    else:
        jump_params_fn=&flat_jump_prob
    if uniform_prior:
        prior_ratio_fn=&uniform_prior_ratio
    else:
        prior_ratio_fn=&flat_prior_ratio
    for v in range(nevents):
        g0[v]=<DTYPE_t>x0[v]['gamma']
        g_s[v]=<DTYPE_t>alpha[v]['gamma']
        d0[v]=<DTYPE_t>x0[v]['delta']
        d_s[v]=<DTYPE_t>alpha[v]['delta']
        h0[v]=<DTYPE_t>x0[v]['h']
        h_s[v]=<DTYPE_t>alpha[v]['h']
        s0[v]=<DTYPE_t>x0[v]['sigma']
        s_s[v]=<DTYPE_t>alpha[v]['sigma']
        k0[v]=<DTYPE_t>x0[v]['kappa']
        if 'proposal_normalisation' in alpha[v].keys():
            proposal_normalisation[v]=<DTYPE_t>alpha[v]['proposal_normalisation']
        if 'gamma_dc' in alpha[v].keys():
            sg[v]=<DTYPE_t>alpha[v]['gamma_dc']
        if 'delta_dc' in alpha[v].keys():
            sd[v]=<DTYPE_t>alpha[v]['delta_dc']
        if dc_prior.shape[0]==1:
            prior_dc[v]=dc_prior[0]
        elif dc_prior.shape[0]>1:
            prior_dc[v]=dc_prior[v]
        for u in range(umax):
            gamma[u,v]=x[v][u]['gamma']
            delta[u,v]=x[v][u]['delta']
            kappa[u,v]=x[v][u]['kappa']
            sigma[u,v]=x[v][u]['sigma']
            h[u,v]=x[v][u]['h']
            if 'g0' in x[v][u].keys():
                qg[u,v]=x[v][u]['g0']#g0 from mt sample ==> dc sample
            else:
                qg[u,v]=x[v][u]['gamma'];#no g0  ==>  mt sample
            if 'd0' in x[v][u].keys():
                qd[u,v]=x[v][u]['d0']
            else:
                qd[u,v]=x[v][u]['delta'];
    index=c_me_acceptance_check(transition_ratio_fn,prior_ratio_fn,jump_params_fn, gamma, delta, h, sigma,kappa, g0, g_s, d0, d_s, h0, h_s, s0, s_s,ln_p,ln_p0,k0,qg,qd,sg,sd,proposal_normalisation,prior_dc)
    returnx=[]
    for v in range(nevents):
        if index>=0:
            returnx.append(x[v][index])
        else:
            returnx.append({})
    if index>=0:
        return returnx,ln_p[index],index
    return returnx,False,umax

#Python function for new sample
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def new_samples(dict x0,dict alpha,Py_ssize_t number_samples, bool dc=False,DTYPE_t jump=0.0,bool gaussian_jump=True):
    cdef Py_ssize_t u
    cdef list x=[]
    cdef DTYPE_t zero=0.0
    cdef dict _x={'gamma':zero,'delta':zero,'kappa':zero,'h':zero,'sigma':zero} 
    cdef DTYPE_t[:,::1] M=np.empty((6,number_samples))
    for u in range(number_samples):
        _x={'gamma':zero,'delta':zero,'kappa':zero,'h':zero,'sigma':zero} 
        if jump and np.random.rand()<jump:
            _x['kappa']=x0['kappa']
            _x['h']=x0['h']
            _x['sigma']=x0['sigma']
            if x0['gamma']==0.0 and x0['delta']==0.0:
                if gaussian_jump:
                    _x['gamma'],_x['delta']=gaussian_jump_params(alpha['gamma_dc'],alpha['delta_dc'])
                else:
                    _x['gamma'],_x['delta']=flat_jump_params(0,0)
            else:
                _x['gamma']=0.0
                _x['delta']=0.0
                _x['g0']=x0['gamma']
                _x['d0']=x0['delta']
        else:
            if dc or x0['gamma']==0.0 and x0['delta']==0.0:
                _x['gamma']=0.0
                _x['delta']=0.0
            else:
                _x['gamma']=new_gamma(<DTYPE_t>x0['gamma'],<DTYPE_t>alpha['gamma'])
                _x['delta']=new_delta(<DTYPE_t>x0['delta'],<DTYPE_t>alpha['delta'])
            _x['kappa']=new_kappa(<DTYPE_t>x0['kappa'],<DTYPE_t>alpha['kappa'])
            _x['h']=new_h(<DTYPE_t>x0['h'],<DTYPE_t>alpha['h'])
            _x['sigma']=new_sigma(<DTYPE_t>x0['sigma'],<DTYPE_t>alpha['sigma'])
        x.append(_x.copy())
        m=convert_sample( _x['gamma'], _x['delta'], _x['kappa'], _x['h'], _x['sigma'])
        M[0,u]=m[0,0]
        M[1,u]=m[1,0]
        M[2,u]=m[2,0]
        M[3,u]=m[3,0]
        M[4,u]=m[4,0]
        M[5,u]=m[5,0]
    return x,np.asarray(M)

# Test functions - Not Documented


class _cmarkov_chain_monte_carlo_TestCase(unittest.TestCase):
    def test_transition_ratios(self):
        self.assertAlmostEqual(gaussian_transition_ratio(0.2,0.1,0.3,0.2,0.2,0.5,0.05,0.7,0.3,0.2,0.1,0.1),1.0011,4)
        self.assertAlmostEqual(gaussian_transition_ratio(0.2,0.05,0.3,0.1, 0.2,0.5, 0.1,0.7, 0.3,0.2, 0.2,0.1),0.9989,4)
        self.assertAlmostEqual(gaussian_transition_ratio(0.1,-0.5,0.2,0.6, 0.05,0.1,0.02,0.1, 0.6,0.2,0.1,0.1),1.1600,4)
    def test_prior_ratios(self):
        self.assertEqual(flat_prior_ratio(0,0,0,0),1.0)
        self.assertEqual(flat_prior_ratio(0,21,0,0),1.0)
        self.assertEqual(flat_prior_ratio(0,0,0,12),1.0)
        self.assertEqual(flat_prior_ratio(0,0,2231,0),1.0)
        self.assertEqual(uniform_prior_ratio(0,0,0,0),1.0)
        self.assertAlmostEqual(uniform_prior_ratio(0,0,0.2,0.2),0.9382,3)
        self.assertAlmostEqual(uniform_prior_ratio(0.2,0.2,0,0),1.0659,3)
        self.assertAlmostEqual(uniform_prior_ratio(0,0,0.2,0.5),1.44127,3)
        self.assertAlmostEqual(uniform_prior_ratio(0.3,0.1,0.2,0.5),1.226,3)
    def test_erf(self):
        import sys
        if 'win' in sys.platform:
            self.assertAlmostEqual(erf(2),0.9953,4)
            self.assertAlmostEqual(erf(0.01),0.0113,4)
            self.assertAlmostEqual(erf(-0.01),-0.0113,4)
            self.assertAlmostEqual(erf(5),1.0000,4)
    def test_convert_sample(self):
        M=convert_sample(0.0,0.0,0.5,0.5,0.0)
        self.assertTrue(M.flags['C_CONTIGUOUS'])
        self.assertEqual(M.shape[0],6)
        self.assertEqual(M.shape[1],1)
    def test_c_acceptance_check(self):
        x=[{'gamma':0.1,'delta':0.05,'h':0.4,'kappa':0.2,'sigma':-0.01},{'gamma':0.05,'delta':0.01,'h':0.45,'kappa':0.4,'sigma':-0.01}]
        ln_p=np.array([0.,50.])
        cdef Py_ssize_t umax=ln_p.shape[0]
        cdef Py_ssize_t u
        cdef DTYPE_t[::1] gamma=np.empty((umax))
        cdef DTYPE_t[::1] delta=np.empty((umax))
        cdef DTYPE_t[::1] h=np.empty((umax))
        cdef DTYPE_t[::1] sigma=np.empty((umax))
        cdef DTYPE_t[::1] kappa=np.empty((umax))   
        cdef DTYPE_t[::1] qg=np.empty((umax))
        cdef DTYPE_t[::1] qd=np.empty((umax))          
        cdef transition_ratio_ptr transition_ratio_fn=&gaussian_transition_ratio
        cdef prior_ratio_ptr prior_ratio_fn=&uniform_prior_ratio  
        cdef jump_params_ptr jump_params_fn=&gaussian_jump_prob
        cdef DTYPE_t g0=0.1
        cdef DTYPE_t g_s=0.1
        cdef DTYPE_t d0=0.05
        cdef DTYPE_t d_s=0.1
        cdef DTYPE_t h0=0.5
        cdef DTYPE_t h_s=0.1
        cdef DTYPE_t s0=0
        cdef DTYPE_t s_s=0.1
        cdef DTYPE_t k0=0.3
        cdef DTYPE_t qg0=0.
        cdef DTYPE_t qd0=0.
        cdef DTYPE_t sg=0.1
        cdef DTYPE_t sd=0.1
        cdef int index
        for u in range(umax):
            gamma[u]=x[u]['gamma']
            delta[u]=x[u]['delta']
            kappa[u]=x[u]['kappa']
            sigma[u]=x[u]['sigma']
            h[u]=x[u]['h']
            qg[u]=0.0;
            qd[u]=0.0;

        index=c_acceptance_check(transition_ratio_fn,prior_ratio_fn,jump_params_fn, gamma, delta, h, sigma,kappa, g0, g_s, d0, d_s, h0, h_s, s0, s_s,ln_p,15.,k0,qg,qd,sg,sd,1,0.5)
        self.assertEqual(index,1)
    def test_gaussian_jump_params(self):
        [g,d]=gaussian_jump_params(0.01,0.01)
        self.assertTrue(fabs(g)<pi/6)
        self.assertTrue(fabs(d)<pi/2)
    def test_flat_jump_params(self):
        [g,d]=flat_jump_params(0.01,0.01)
        self.assertTrue(fabs(g)<pi/6)
        self.assertTrue(fabs(d)<pi/2)
    def test_flat_jump_prob(self):
        p=flat_jump_prob(0,0,1,1,1)
        self.assertEqual(p,3/(PI2))
        p=flat_jump_prob(3,1,2,4,1)
        self.assertEqual(p,3/(PI2))
    def test_gaussian_jump_prob(self):
        p=gaussian_jump_prob(0,0,1,1,1)
        self.assertEqual(p,gaussian_pdf(0,0,1)*gaussian_pdf(0,0,1))
        p=gaussian_jump_prob(0.2,0.2,0.2,0.2,2)
        self.assertEqual(p,gaussian_pdf(0.2,0,0.2)*gaussian_pdf(0.2,0,0.2)/2)
    def test_new_samples(self):
        x0={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': 0.0, 'delta': 0.0}
        alpha={'gamma_dc': 0.10471975511965977, 'sigma_dc': 0.009085556327826677, 'kappa': 0.018171112655653354, 'delta_dc': 0.15707963267948966,
         'h': 0.005784044801253858, 'h_dc': 0.005784044801253858, 'poisson': 0.005784044801253858, 'delta': 0.009085556327826677,
          'alpha': 0.009085556327826677, 'kappa_dc': 0.018171112655653354, 'sigma': 0.009085556327826677, 'gamma': 0.00605703755188445,'proposal_normalisation':1.0}
        x,M=new_samples(x0,alpha,2)
        self.assertEqual(len(x),2)
        self.assertEqual(M.shape[0],6)
        self.assertEqual(M.shape[1],2)
        self.assertTrue(M.flags['C_CONTIGUOUS'])
        x,M=new_samples(x0,alpha,1)
        self.assertEqual(len(x),1)
        self.assertEqual(M.shape[0],6)
        self.assertEqual(M.shape[1],1)
        self.assertTrue(M.flags['C_CONTIGUOUS'])
        x,M=new_samples(x0,alpha,1,True,1)
        self.assertEqual(len(x),1)
        self.assertEqual(M.shape[0],6)
        self.assertEqual(M.shape[1],1)
        self.assertTrue('gamma' in x[0].keys())
        self.assertFalse('g0' in x[0].keys())
        x0={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': 0.5, 'delta': 0.5}
        x,M=new_samples(x0,alpha,1,False,1)
        self.assertTrue('gamma' in x[0].keys())
        self.assertTrue('g0' in x[0].keys())
        self.assertTrue(M.flags['C_CONTIGUOUS'])
    def test_acceptance_check(self):
        x0={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': 0.0, 'delta': 0.0}
        x1={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': .3, 'delta': 0.5}
        alpha={'gamma_dc': 0.10471975511965977, 'sigma_dc': 0.009085556327826677, 'kappa': 0.018171112655653354, 'delta_dc': 0.15707963267948966,
         'h': 0.005784044801253858, 'h_dc': 0.005784044801253858, 'poisson': 0.005784044801253858, 'delta': 0.009085556327826677,
          'alpha': 0.009085556327826677, 'kappa_dc': 0.018171112655653354, 'sigma': 0.009085556327826677, 'gamma': 0.00605703755188445,'proposal_normalisation':1.0}
        ln_p=np.array([5.0,3.0])
        ln_p0=3.0
        x=[x0,x1]
        xi,ln_pi,i=acceptance_check(x,x0,alpha,ln_p,ln_p0,True,False)
        self.assertEqual(i,0) 
        self.assertEqual(x0,xi)
        ln_p=np.array([-np.inf,3.0])
        ln_p0=-np.inf
        x=[x0,x1]
        xi,ln_pi,i=acceptance_check(x,x0,alpha,ln_p,ln_p0,True,False)
        self.assertEqual(i,1)
        self.assertEqual(x1,xi)
        #Tests for gaussian jump
        x0={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': 0.0, 'delta': 0.0}
        x1={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': .1, 'delta': 0.1}
        alpha={'gamma_dc': 0.1, 'sigma_dc': 0.009085556327826677, 'kappa': 0.018171112655653354, 'delta_dc': 0.1,
         'h': 0.005784044801253858, 'h_dc': 0.005784044801253858, 'poisson': 0.005784044801253858, 'delta': 0.009085556327826677,
          'alpha': 0.009085556327826677, 'kappa_dc': 0.018171112655653354, 'sigma': 0.009085556327826677, 'gamma': 0.00605703755188445,'proposal_normalisation':0.99501231590631156}
        ln_p=np.array([0.])
        ln_p0=np.log(np.array([0.231]))
        #dc to mt
        xi,ln_pi,i=acceptance_check([x1],x0,alpha,ln_p,ln_p0,True,True)
        self.assertTrue(len(xi)>0)
        ln_p=np.array([-10000000000000000000000.])
        ln_p0=np.array([100000000000000000000000000000000000000000000])
        #dc to mt
        xi,ln_pi,i=acceptance_check([x1],x0,alpha,ln_p,ln_p0,True,True)
        self.assertEqual(len(xi),0)
        ln_p=np.log(np.array([0.2321]))
        ln_p0=np.array([0.])
        #mt to dc
        ln_p=np.array([0.2329])
        xi,ln_pi,i=acceptance_check([x0],x1,alpha,ln_p,ln_p0,True,True)
        self.assertTrue(len(xi)>0)
        ln_p=np.array([-10000000000000000000000.])
        ln_p0=np.array([100000000000000000000000000000000000000000000.])
        xi,ln_pi,i=acceptance_check([x0],x1,alpha,ln_p,ln_p0,True,True)
        self.assertEqual(len(xi),0)
    def test_c_me_acceptance_check(self):
        x=[{'gamma':0.1,'delta':0.05,'h':0.4,'kappa':0.2,'sigma':-0.01},{'gamma':0.05,'delta':0.01,'h':0.45,'kappa':0.4,'sigma':-0.01}]
        ln_p=np.array([0.,10000])
        cdef Py_ssize_t umax=ln_p.shape[0]
        cdef Py_ssize_t nevents=2
        cdef Py_ssize_t u
        cdef Py_ssize_t v
        cdef DTYPE_t[:,::1] gamma=np.empty((umax,nevents))
        cdef DTYPE_t[:,::1] delta=np.empty((umax,nevents))
        cdef DTYPE_t[:,::1] h=np.empty((umax,nevents))
        cdef DTYPE_t[:,::1] sigma=np.empty((umax,nevents))
        cdef DTYPE_t[:,::1] kappa=np.empty((umax,nevents))   
        cdef DTYPE_t[:,::1] qg=np.empty((umax,nevents))
        cdef DTYPE_t[:,::1] qd=np.empty((umax,nevents))          
        cdef transition_ratio_ptr transition_ratio_fn=&gaussian_transition_ratio
        cdef prior_ratio_ptr prior_ratio_fn=&uniform_prior_ratio  
        cdef jump_params_ptr jump_params_fn=&gaussian_jump_prob
        cdef DTYPE_t [::1]g0=np.empty((nevents))#0.1
        cdef DTYPE_t [::1]g_s=np.empty((nevents))#0.1
        cdef DTYPE_t [::1]d0=np.empty((nevents))#0.05
        cdef DTYPE_t [::1]d_s=np.empty((nevents))#0.1
        cdef DTYPE_t [::1]h0=np.empty((nevents))#0.5
        cdef DTYPE_t [::1]h_s=np.empty((nevents))#0.1
        cdef DTYPE_t [::1]s0=np.empty((nevents))#0
        cdef DTYPE_t [::1]s_s=np.empty((nevents))#0.1
        cdef DTYPE_t [::1]k0=np.empty((nevents))#0.3
        cdef DTYPE_t [::1]qg0=np.empty((nevents))#0.
        cdef DTYPE_t [::1]qd0=np.empty((nevents))#0.
        cdef DTYPE_t [::1]sg=np.empty((nevents))#0.1
        cdef DTYPE_t [::1]sd=np.empty((nevents))#0.1
        cdef DTYPE_t [::1]pn=np.empty((nevents))#0.1
        cdef DTYPE_t [::1]dc_prior=np.empty((nevents))#0.1
        cdef int index
        for v in range(nevents):
            g0[v]=0.1
            g_s[v]=0.1
            d0[v]=0.05
            d_s[v]=0.1
            h0[v]=0.5
            h_s[v]=0.1
            s0[v]=0.
            s_s[v]=0.1
            k0[v]=0.3
            qg0[v]=0.
            qd0[v]=0.
            sg[v]=0.1
            sd[v]=0.1
            pn[v]=1
            dc_prior[v]=0.5
            for u in range(umax):
                gamma[u,v]=x[u]['gamma']
                delta[u,v]=x[u]['delta']
                kappa[u,v]=x[u]['kappa']
                sigma[u,v]=x[u]['sigma']
                h[u,v]=x[u]['h']
                qg[u,v]=0.0;
                qd[u,v]=0.0;
        index=c_me_acceptance_check(transition_ratio_fn,prior_ratio_fn,jump_params_fn, gamma, delta, h, sigma,kappa, g0, g_s, d0, d_s, h0, h_s, s0, s_s,ln_p,-200.,k0,qg,qd,sg,sd,pn,dc_prior)
        self.assertEqual(index,0)
        index=c_me_acceptance_check(transition_ratio_fn,prior_ratio_fn,jump_params_fn, gamma, delta, h, sigma,kappa, g0, g_s, d0, d_s, h0, h_s, s0, s_s,ln_p,100,k0,qg,qd,sg,sd,pn,dc_prior)
        self.assertEqual(index,1)

    def test_me_acceptance_check(self):
        x0={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': 0.0, 'delta': 0.0}
        x1={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': .3, 'delta': 0.5}
        alpha={'gamma_dc': 0.10471975511965977, 'sigma_dc': 0.009085556327826677, 'kappa': 0.018171112655653354, 'delta_dc': 0.15707963267948966,
         'h': 0.005784044801253858, 'h_dc': 0.005784044801253858, 'poisson': 0.005784044801253858, 'delta': 0.009085556327826677,
          'alpha': 0.009085556327826677, 'kappa_dc': 0.018171112655653354, 'sigma': 0.009085556327826677, 'gamma': 0.00605703755188445,'proposal_normalisation':1.0}
        ln_p=np.array([5.0,3.0])
        ln_p0=-30
        x=[x0,x1]
        xi,ln_pi,i=me_acceptance_check([x,x,x],[x0,x0,x0],[alpha,alpha,alpha],ln_p,ln_p0,True,False)
        self.assertEqual(i,0) 
        self.assertEqual([x0,x0,x0],xi)
        ln_p=np.array([-np.inf,3.0])
        ln_p0=-np.inf
        x=[x0,x1]
        xi,ln_pi,i=me_acceptance_check([x],[x0],[alpha],ln_p,ln_p0,True,False)
        self.assertEqual(i,1)
        self.assertEqual([x1],xi)
        #Tests for gaussian jump
        x0={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': 0.0, 'delta': 0.0}
        x1={'h': 0.16037522711021507, 'kappa': 1.7045926666235598, 'sigma': 1.2202710645123613, 'gamma': .1, 'delta': 0.1}
        alpha={'gamma_dc': 0.1, 'sigma_dc': 0.009085556327826677, 'kappa': 0.018171112655653354, 'delta_dc': 0.1,
         'h': 0.005784044801253858, 'h_dc': 0.005784044801253858, 'poisson': 0.005784044801253858, 'delta': 0.009085556327826677,
          'alpha': 0.009085556327826677, 'kappa_dc': 0.018171112655653354, 'sigma': 0.009085556327826677, 'gamma': 0.00605703755188445,'proposal_normalisation':0.99501231590631156}
        ln_p=np.array([0.])
        ln_p0=np.log(np.array([0.231]))
        #dc to mt
        xi,ln_pi,i=me_acceptance_check([[x1,x1]],[x0],[alpha],ln_p,ln_p0,True,True)
        self.assertTrue(len(xi)>0)
        ln_p=np.array([-10000000000000000000000.])
        ln_p0=np.array([100000000000000000000000000000000000000000000])
        #dc to mt
        xi,ln_pi,i=me_acceptance_check([[x1]],[x0],[alpha],ln_p,ln_p0,True,True)
        self.assertEqual(len(xi[0]),0)
        ln_p=np.log(np.array([0.2321]))
        ln_p0=np.array([0.])
        #mt to dc
        ln_p=np.array([0.2329])
        xi,ln_pi,i=me_acceptance_check([[x0]],[x1],[alpha],ln_p,ln_p0,True,True)
        self.assertTrue(len(xi[0])>0)
        ln_p=np.array([-10000000000000000000000.])
        ln_p0=np.array([100000000000000000000000000000000000000000000.])
        xi,ln_pi,i=me_acceptance_check([[x0]],[x1],[alpha],ln_p,ln_p0,True,True)
        self.assertEqual(len(xi[0]),0)


def _acceptance_test_fn(dict x, dict x0, dict alpha, DTYPE_t ln_p,DTYPE_t ln_p0,bool uniform_prior=True,bool gaussian_jump=True,DTYPE_t p_dc=0.5):
    cdef DTYPE_t qg
    cdef DTYPE_t qd
    cdef transition_ratio_ptr transition_ratio_fn=&gaussian_transition_ratio
    cdef prior_ratio_ptr prior_ratio_fn    
    cdef jump_params_ptr jump_params_fn
    cdef DTYPE_t proposal_normalisation=1.0
    if 'proposal_normalisation' in alpha.keys():
        proposal_normalisation=<DTYPE_t>alpha['proposal_normalisation']
    cdef DTYPE_t sg=0.1
    cdef DTYPE_t sd=0.1
    cdef int index
    if gaussian_jump:
        jump_params_fn=&gaussian_jump_prob
    else:
        jump_params_fn=&flat_jump_prob
    if uniform_prior:
        prior_ratio_fn=&uniform_prior_ratio
    else:
        prior_ratio_fn=&flat_prior_ratio
    if 'gamma_dc' in alpha.keys():
        sg=<DTYPE_t>alpha['gamma_dc']
    if 'delta_dc' in alpha.keys():
        sd=<DTYPE_t>alpha['delta_dc']
    if 'g0' in x.keys():
        qg=x['g0']#g0 from mt sample ==> dc sample
    else:
        qg=x['gamma'];#no g0  ==>  mt sample
    if 'd0' in x.keys():
        qd=x['d0']
    else:
        qd=x['delta'];
    cdef int jump=0
    if ((x0['gamma']==0.0 and x0['delta']==0.0) or (x['gamma']==0.0 and x['delta']==0.0)) and  not x['gamma']==x0['gamma'] and not x['delta']==x0['delta'] and x0['h']==x['h'] and x0['sigma']==x['sigma'] and x0['kappa']==x['kappa']:
        jump=1
    if ln_p==-inf:
        return 0
    return acceptance(transition_ratio_fn,prior_ratio_fn,jump_params_fn,x['gamma'], x['delta'], x['h'], x['sigma'], x0['gamma'], alpha['gamma'], x0['delta'], alpha['delta'], x0['h'], alpha['h'], x0['sigma'], alpha['sigma'], ln_p, ln_p0,jump,qg,qd,sg,sd,proposal_normalisation,p_dc)

def _gaussian_cdf_test(DTYPE_t x,DTYPE_t mu,DTYPE_t s):
    return gaussian_cdf(x,mu,s)

def _gaussian_transition_ratio_test(dict x, dict x0,dict alpha):
    return gaussian_transition_ratio(x['gamma'], x['delta'], x['h'], x['sigma'], x0['gamma'], alpha['gamma'], x0['delta'], alpha['delta'], x0['h'], alpha['h'], x0['sigma'], alpha['sigma']) 

def _test_suite():
    _c_markov_chain_monte_carlo_test_suite = unittest.TestLoader().loadTestsFromTestCase(_cmarkov_chain_monte_carlo_TestCase)
    return unittest.TestSuite([_c_markov_chain_monte_carlo_test_suite])

def run_tests():
    """Runs monteCarloMarkovChain module tests.
    """
    suite=_test_suite()
    unittest.TextTestRunner(verbosity=4).run(suite)

if  __name__=='__main__':
    run_tests()


"""
test_multiprocessing_utils.py
*****************************

Tests for src/utils/multiprocessing_utils.py
"""
import sys
import unittest
import multiprocessing
import multiprocessing.queues


from MTfit.utilities.unittest_utils import TestCase
from MTfit.utilities.multiprocessing_helper import Worker
from MTfit.utilities.multiprocessing_helper import JobPool
from MTfit.utilities.multiprocessing_helper import PoisonPill
from MTfit.tests.unit.utilities.multiprocessing_test_classes import TaskTest
from MTfit.tests.unit.utilities.multiprocessing_test_classes import TaskTest2


class WorkerTestCase(TestCase):

    def setUp(self):
        self.task_queue = multiprocessing.Queue()
        self.result_queue = multiprocessing.Queue()
        self.worker = Worker(self.task_queue, self.result_queue)
        self.worker.start()

    def tearDown(self):
        self.task_queue.put(PoisonPill())
        if self.worker.is_alive():
            self.worker.join()
        del self.worker
        self.task_queue.close()
        del self.task_queue
        self.result_queue.close()
        del self.result_queue

    def test___init__(self):
        self.assertEqual(self.worker.task_queue, self.task_queue)
        self.assertEqual(self.worker.result_queue, self.result_queue)
        self.assertTrue(
            isinstance(self.worker.task_queue, multiprocessing.queues.Queue))
        self.assertTrue(
            isinstance(self.worker.result_queue, multiprocessing.queues.Queue))

    def test_run(self):
        self.test___init__()
        self.assertTrue(self.result_queue.empty())
        self.task_queue.put(TaskTest(1, 2))
        r = self.result_queue.get(timeout=10)
        self.assertTrue(self.result_queue.empty())
        self.assertEqual(r, 2)
        self.task_queue.put(TaskTest(1, 5))
        r = self.result_queue.get(timeout=10)
        self.assertTrue(self.result_queue.empty())
        self.assertEqual(r, 5)


class JobPoolTestCase(TestCase):

    def setUp(self, task=TaskTest, single_life=False):
        self.job_pool = JobPool(2, task=task, single_life=single_life)

    def tearDown(self):
        self.job_pool.close()
        del self.job_pool

    def test_close(self):
        for w in self.job_pool.workers:
            self.assertTrue(w.is_alive())
        self.job_pool.close()
        for w in self.job_pool.workers:
            self.assertFalse(w.is_alive())

    def test_task(self):
        self.job_pool.task(1, 2)
        self.assertEqual(self.job_pool.number_jobs, 1)

    def test_custom_task(self):
        self.job_pool.custom_task(TaskTest2, 4, 2)
        self.assertEqual(self.job_pool.number_jobs, 1)
        self.assertEqual(self.job_pool.result(), 6)

    def test_result(self):
        self.job_pool.task(1, 2)
        result = self.job_pool.result()
        self.assertEqual(result, 2)

    def test_all_results(self):
        self.job_pool.task(1, 2)
        self.job_pool.task(1, 3)
        self.assertEqual(sorted(self.job_pool.all_results()), [2, 3])

    def test_parallel(self):
        self.job_pool.custom_task(TaskTest2, 2, 4)
        self.job_pool.custom_task(TaskTest2, 4, 1)
        self.job_pool.custom_task(TaskTest2, 4, 0)
        self.assertNotEqual(self.job_pool.all_results(), [6, 5, 4])

    def test_single_life(self):
        self.tearDown()
        self.setUp(single_life=True)
        self.assertEqual(self.job_pool.number_workers, 2)
        self.job_pool.custom_task(TaskTest2, 2, 4)
        self.job_pool.custom_task(TaskTest2, 4, 1)
        self.job_pool.custom_task(TaskTest2, 4, 0)
        self.job_pool.custom_task(TaskTest2, 8, 1)
        self.assertNotEqual(self.job_pool.all_results(), [6, 5, 4, 9])

    def test_exceptionBreak(self):
        self.job_pool.custom_task(TaskTest2, 2, 4)
        self.job_pool.custom_task(TaskTest2, "4", 1)
        self.job_pool.custom_task(TaskTest2, "4", 0)
        self.job_pool.custom_task(TaskTest2, "8", 1)
        results = self.job_pool.all_results()
        self.assertTrue(6 in results)
        results.pop(results.index(6))
        self.assertTrue(isinstance(results[0], TypeError))



"""
test_unittest_utils.py
**********************

Tests for src/utils/unittest_utils.py
"""

import numpy as np

from MTfit.utilities.unittest_utils import TestCase


class TestCaseTestCase(TestCase):

    def test_assertAlmostEquals(self):
        self.assertAlmostEquals(1.000000, 0.9999999999)
        self.assertAlmostEquals(np.array([1.000000]), np.array([0.9999999999]))
        self.assertAlmostEquals(np.array([1.000000]), [0.9999999999])
        self.assertAlmostEquals({'x': 1.000000}, {'x': 0.9999999999})

    def test_assertEquals(self):
        self.assertEquals(1.000000, 1.000000)
        self.assertEquals(np.array([1.000000]), np.array([1.000000]))
        self.assertEquals(np.array([1.000000]), [1.000000])
        self.assertEquals({'x': 1.000000}, {'x': 1.000000})



# cython: infer_types=True
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION

# **Restricted:  For Non-Commercial Use Only** 
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.

cimport cython
import numpy as np
cimport numpy as np
from cython.view cimport array as cvarray
from cpython cimport bool
# DTYPE=np.float64
# ctypedef np.float64_t DTYPE_t
ctypedef double DTYPE_t
ctypedef long long LONG
# ctypedef long long
from libc.stdlib cimport rand, RAND_MAX
IF UNAME_SYSNAME == "Windows":
    from libc.math cimport HUGE_VAL as inf
ELSE:
    from libc.math cimport INFINITY as inf
    from libc.math cimport fmax
    from libc.math cimport fmin
    from libc.math cimport erf
from libc.math cimport sqrt
from libc.math cimport exp
from libc.math cimport log
from libc.math cimport fabs
from libc.math cimport cos
from libc.math cimport sin
from libc.math cimport M_PI as pi
from libc.math cimport M_SQRT2 as sqrt2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef DTYPE_t[::1] get_multipliers(DTYPE_t[:,:,::1] angles,DTYPE_t bin_size,DTYPE_t [::1] multipliers) nogil:
    cdef Py_ssize_t umax=angles.shape[0]
    cdef Py_ssize_t nsta=angles.shape[2]
    cdef Py_ssize_t u
    cdef Py_ssize_t w
    cdef int ok=1
    for u in range(umax):
        if multipliers[u]==-1:
            continue
        for v in range(u+1,umax):  
            if multipliers[v]>-1:
                ok=1
                for w in range(nsta):
                    ok=ok*(fabs(angles[u,0,w]-angles[v,0,w])<bin_size/2.)*(fabs(angles[u,1,w]-angles[v,1,w])<bin_size/2.)
                    if ok==0:
                        break
                if w==nsta-1 and ok>0:
                    multipliers[u]+=multipliers[v]
                    if v>u:
                        multipliers[v]=-1
    return multipliers


cpdef bin_scatangle(sample_records,multipliers,bin_size):
    cdef DTYPE_t[:,:,::1] angles=np.empty((len(sample_records),2,len(sample_records[0]['Name'])))
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    new_records=[]
    new_multipliers=[]
    for i,record in enumerate(sample_records):
        # print i,record
        for j in range(len(record['Name'])):
            angles[i,0,j]=record['TakeOffAngle'][j,0]
            angles[i,1,j]=record['Azimuth'][j,0]
    # print multipliers,bin_size,angles
    multipliers=np.asarray(get_multipliers(angles,bin_size,multipliers)).flatten()
    for i,record in enumerate(sample_records):
        if multipliers[i]>0:
            new_records.append(record)
            new_multipliers.append(multipliers[i])
    return new_records,new_multipliers


"""
spherical_projection
********************

Spherical projection functions:
    * Equal Area
    * Equal Angle
"""


# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.


import numpy as np


def equal_area(x, y, z, projection_axis=False, lower=True, full_sphere=False, back_project=False):
    """
    Equal area projection of cartesian points on a sphere to cartesian points on a plane.

    Args:
        x: array-like - array of spherical cartesian coordinates to transform
        y: array-like - array of spherical cartesian coordinates to transform
        z: array-like - array of spherical cartesian coordinates to transform


    Keyword Args:
        projection_axis: array-like - vector of the projection axis
        lower: boolean - boolean flag for lower or upper hemisphere projection
        full_sphere: boolean - boolean flag for full sphere projection
        back_project: boolean - boolean flag for back projecting the results

    Returns:
        (np.array, np.array) - arrays of corresponding projected X and Y coordinates
    """

    def equal_area_correction(x, y, z, lower=True):
        if lower:
            return np.sqrt(2/(1+z))
        return np.sqrt(2/(1-z))
    return __project(x, y, z, equal_area_correction, projection_axis, lower, full_sphere, back_project)


def equal_angle(x, y, z, projection_axis=False, lower=True, full_sphere=False, back_project=False):
    """
    Equal angle (stereographic) projection of cartesian points on a sphere to cartesian points on a plane.

    Args:
        x: array-like - array of spherical cartesian coordinates to transform
        y: array-like - array of spherical cartesian coordinates to transform
        z: array-like - array of spherical cartesian coordinates to transform


    Keyword Args:
        projection_axis: array-like - vector of the projection axis
        lower: boolean - boolean flag for lower or upper hemisphere projection
        full_sphere: boolean - boolean flag for full sphere projection
        back_project: boolean - boolean flag for back projecting the results

    Returns:
        (np.array, np.array) - arrays of corresponding projected X and Y coordinates
    """

    def equal_angle_correction(x, y, z, lower=True):
        if lower:
            return 1/(1+z)
        return 1/(1-z)
    return __project(x, y, z, equal_angle_correction, projection_axis, lower, full_sphere, back_project)


def __project(x, y, z, correction_function, projection_axis=False, lower=True, full_sphere=False, back_project=False):
    """
    Helper function for projection of cartesian points on a sphere to cartesian points on a plane.

    Args:
        x: array-like - array of spherical cartesian coordinates to transform
        y: array-like - array of spherical cartesian coordinates to transform
        z: array-like - array of spherical cartesian coordinates to transform
        correction_function: scaling correction based on the projection

    Keyword Args:
        projection_axis: array-like - vector of the projection axis
        lower: boolean - boolean flag for lower or upper hemisphere projection
        full_sphere: boolean - boolean flag for full sphere projection
        back_project: boolean - boolean flag for back projecting the results

    Returns:
        (np.array, np.array) - arrays of corresponding projected X and Y coordinates
    """

    if not isinstance(x, np.ndarray):
        x = np.array([x])
    if not isinstance(y, np.ndarray):
        y = np.array([y])
    if not isinstance(z, np.ndarray):
        z = np.array([z])
    if not len(x.shape):
        x = np.array([x])
    if not len(y.shape):
        y = np.array([y])
    if not len(z.shape):
        z = np.array([z])
    x = np.array(x)
    y = np.array(y)
    z = np.array(z)
    if isinstance(projection_axis, (list, np.ndarray)):
        projection_axis = np.array(projection_axis).flatten()
        rAxis = np.cross(projection_axis, np.array([0, 0, 1]))
        theta = - np.arcsin(np.sqrt(np.sum(rAxis ^ 2)) /
                            (np.sqrt(np.sum(projection_axis ^ 2))))
        rAxis = rAxis/np.sqrt(np.sum(rAxis ^ 2))
        R = np.matrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])*np.cos(theta) + \
            np.sin(theta)*np.matrix([[0, -rAxis[2], rAxis[1]],
                                     [rAxis[2], 0, -rAxis[0]],
                                     [-rAxis[1], rAxis[0], 0]]) + \
            (1-np.cos(theta))*(np.matrix(rAxis).T*np.matrix(rAxis))
        V = R*np.matrix([[x], [y], [z]])
        x = V[1, :]
        y = V[2, :]
        z = V[3, :]
    X = x*correction_function(x, y, z, lower)
    Y = y*correction_function(x, y, z, lower)
    if not full_sphere:
        if lower:
            X[z < 0] = np.nan
            Y[z < 0] = np.nan
        else:
            X[z > 0] = np.nan
            Y[z > 0] = np.nan
    if back_project:
        X_b = -x*correction_function(-x, -y, -z, lower)
        Y_b = -y*correction_function(-x, -y, -z, lower)
        X[np.isnan(X)] = X_b[np.isnan(Y)]
        Y[np.isnan(Y)] = Y_b[np.isnan(Y)]
    X[np.isinf(X)] = np.nan
    Y[np.isinf(Y)] = np.nan
    return X, Y



import os


def simple_parser(filename):
    # One file per event
    lines = open(filename).readlines()
    # Set up polarity data dictionary
    polarity_dict = {'Stations': {'Name': [], 'TakeOffAngle': [], 'Azimuth': []}, 'Measured': [], 'Error': []}
    # Loop over lines
    for l in lines:
        # Check there are 5 data types
        if len(l.split('\t')) == 5:
            # Get data from line
            name, polarity, error, azimuth, takeoff_angle = l.rstrip().split('\t')
            # Append data to the polarity_dict (and convert to correct format)
            polarity_dict['Stations']['Name'].append(name)
            polarity_dict['Stations']['TakeOffAngle'].append(float(takeoff_angle))
            polarity_dict['Stations']['Azimuth'].append(float(azimuth))
            polarity_dict['Measured'].append(float(polarity))
            polarity_dict['Error'].append(float(error))
    # filename as UID
    event_dict = {'UID': os.path.splitext(os.path.split(filename)[-1])[0], 'PPolarity': polarity_dict}
    # Return list of event dictionary
    return [event_dict]



#!python
# cython: infer_types=True

# **Restricted:  For Non-Commercial Use Only** 
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.

from libc.math cimport fabs
from libc.math cimport fmod
from libc.math cimport cos
from libc.math cimport acos
from libc.math cimport atan2
from libc.math cimport sin
from libc.math cimport sqrt
from libc.math cimport M_PI as pi
from libc.math cimport M_SQRT2 as sqrt2
from libc.stdlib cimport free

import unittest

cimport cython
cimport numpy as np
from scipy.linalg import eigh
from scipy.optimize import fsolve
from scipy import __version__ as __scipy_version__
import numpy as np
from cpython cimport bool

from ..utilities.unittest_utils import TestCase


cdef DTYPE_t PI2=2*pi
cdef DTYPE_t sqrt3=sqrt(3)
cdef DTYPE_t rad_cor=180/pi
cdef int check_finite=1
if int(__scipy_version__.split('.')[0])==0 and int(__scipy_version__.split('.')[1])<13:
    check_finite=0
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t[::1] cE_tk(DTYPE_t[::1] E,DTYPE_t[::1] results) nogil:
    cdef DTYPE_t iso=(E[0]+E[1]+E[2])/3
    cdef DTYPE_t dev0=E[0]-iso
    cdef DTYPE_t dev1=E[2]-iso#Odd sorting from hudson paper E[0]>=E[2]>=E[1]
    cdef DTYPE_t dev2=E[1]-iso
    if dev2>0:
        results[5]=iso/(fabs(iso)-dev1)
        results[6]=-2*(dev2)/(dev1)
    elif dev2<0:
        results[5]=iso/(fabs(iso)+dev0)
        results[6]=2*dev2/dev0
    else:
        results[5]=iso/(fabs(iso)+dev0)
        results[6]=0
    results[6]=results[6]*(1-fabs(results[5]));
    return results
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline DTYPE_t[::1] ctk_uv(DTYPE_t[::1] results) nogil:
    cdef DTYPE_t k=results[5]
    cdef DTYPE_t tau=results[6]
    if tau>0 and k>0:
        if tau<4*k:
            results[5]=tau/(1-(tau/2))
            results[6]=k/(1-(tau/2))
        else:
            results[5]=tau/(1-2*k)
            results[6]=k/(1-2*k)
    elif tau<0 and k<0:
        if tau>4*k:
            results[5]=tau/(1+(tau/2))
            results[6]=k/(1+(tau/2))
        else:
            results[5]=tau/(1+2*k)
            results[6]=k/(1+2*k)
    else:
        results[5]=tau
        results[6]=k
    return results
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline void cE_gd(DTYPE_t[:] E,DTYPE_t*g,DTYPE_t*d) nogil:
    if E[0]==E[2]:
        if E[0]>0:
            g[0]=0
            d[0]=pi/2
        elif E[0]<0:
            g[0]=0
            d[0]=-pi/2
    else:
        g[0]=atan2(-E[0]+2*E[1]-E[2],sqrt3*(E[0]-E[2]))
        d[0]=pi/2-acos((E[0]+E[1]+E[2])/(sqrt3*sqrt((E[0]*E[0]+E[1]*E[1]+E[2]*E[2]))))
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline void cN_SDR(DTYPE_t N0,DTYPE_t N1,DTYPE_t N2,DTYPE_t S0,DTYPE_t S1,DTYPE_t S2,DTYPE_t* strike,DTYPE_t*dip,DTYPE_t*rake) nogil:
    if N2 > 0:
        N0 = -N0
        N1 = -N1
        N2 = -N2
        S0 = -S0
        S1 = -S1
        S2 = -S2
    strike[0] = atan2(-N0, N1)
    dip[0] = atan2((N1**2+N0**2), sqrt((N0*N2)**2+(N1*N2)**2))
    rake[0] = atan2(-S2, S0*N1 - S1*N0)
    if dip[0] > pi/2:
        dip[0] = pi-dip[0]
        strike[0] = strike[0]+pi
        rake[0] = PI2-rake[0]
    if fabs(strike[0]) > PI2:
        strike[0] = fmod(strike[0],PI2)
    if strike[0] < 0.0:
        strike[0] += PI2
    if rake[0] > pi:
        rake[0] -= PI2
    elif rake[0] < -pi:
        rake[0] += PI2

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef DTYPE_t[::1] cTP_SDR(DTYPE_t[::1] T,DTYPE_t[::1] P,DTYPE_t[::1] results) nogil:    
    cdef DTYPE_t Nt=sqrt((T[0]+P[0])*(T[0]+P[0])+(T[1]+P[1])*(T[1]+P[1])+(T[2]+P[2])*(T[2]+P[2]))
    cdef DTYPE_t St=sqrt((T[0]-P[0])*(T[0]-P[0])+(T[1]-P[1])*(T[1]-P[1])+(T[2]-P[2])*(T[2]-P[2]))
    cdef DTYPE_t N2=(T[2]+P[2])/Nt
    cdef DTYPE_t N0=(T[0]+P[0])/Nt
    cdef DTYPE_t N1=(T[1]+P[1])/Nt
    cdef DTYPE_t S2=(T[2]-P[2])/St
    cdef DTYPE_t S0=(T[0]-P[0])/St
    cdef DTYPE_t S1=(T[1]-P[1])/St
    cdef DTYPE_t strike1=0.
    cdef DTYPE_t dip1=0.
    cdef DTYPE_t rake1=0.
    cdef DTYPE_t strike2=0.
    cdef DTYPE_t dip2=0.
    cdef DTYPE_t rake2=0.
    cN_SDR(N0,N1,N2,S0,S1,S2,&strike1,&dip1,&rake1)
    cN_SDR(S0,S1,S2,N0,N1,N2,&strike2,&dip2,&rake2)
    #Switch for sdr2
    if fabs(rake1)>pi/2:   
        results[2]=strike2
        results[3]=cos(dip2)
        results[4]=rake2
    else:        
        results[2]=strike1
        results[3]=cos(dip1)
        results[4]=rake1
    results[7]=strike1*rad_cor
    results[8]=dip1*rad_cor
    results[9]=rake1*rad_cor
    results[10]=strike2*rad_cor
    results[11]=dip2*rad_cor
    results[12]=rake2*rad_cor
    return results
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef DTYPE_t[::1] cTPE_convert(DTYPE_t[::1] T,DTYPE_t[::1] P,DTYPE_t[::1]E,DTYPE_t[::1]results) nogil: 
    cE_gd(E,&results[0],&results[1])
    results=cE_tk(E,results)
    results=ctk_uv(results)
    results=cTP_SDR(T,P,results)
    return results
#Tape to Moment tensor conversion
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void cTape_MT6(DTYPE_t*M, DTYPE_t gamma,DTYPE_t delta,DTYPE_t kappa,DTYPE_t h,DTYPE_t sigma) nogil:
    #Get T N P axes
    cdef DTYPE_t ck=cos(kappa)
    cdef DTYPE_t cs=cos(sigma)
    cdef DTYPE_t sk=sin(kappa)
    cdef DTYPE_t ss=sin(sigma)
    cdef DTYPE_t cg=cos(gamma)
    cdef DTYPE_t sg=sin(gamma)
    cdef DTYPE_t cd=cos(delta)
    cdef DTYPE_t sh=sqrt(1-h*h)
    cdef DTYPE_t NT=sqrt((ck*cs+sk*h*ss-sk*sh)*(ck*cs+sk*h*ss-sk*sh)+(sk*cs-ck*h*ss+ck*sh)*(sk*cs-ck*h*ss+ck*sh)+(-sh*ss-h)*(-sh*ss-h))
    cdef DTYPE_t NP=sqrt((ck*cs+sk*h*ss+sk*sh)*(ck*cs+sk*h*ss+sk*sh)+(sk*cs-ck*h*ss-ck*sh)*(sk*cs-ck*h*ss-ck*sh)+(-sh*ss+h)*(-sh*ss+h))
    cdef DTYPE_t T1=(ck*cs+sk*h*ss-sk*sh)/NT
    cdef DTYPE_t T2=(sk*cs-ck*h*ss+ck*sh)/NT
    cdef DTYPE_t T3=(-sh*ss-h)/NT
    cdef DTYPE_t P1=(ck*cs+sk*h*ss+sk*sh)/NP
    cdef DTYPE_t P2=(sk*cs-ck*h*ss-ck*sh)/NP
    cdef DTYPE_t P3=(-sh*ss+h)/NP
    cdef DTYPE_t N1=T2*P3-P2*T3
    cdef DTYPE_t N2=-T1*P3+P1*T3
    cdef DTYPE_t N3=T1*P2-T2*P1
    #Normalised TNP Values
    cdef DTYPE_t sd=sin(delta)
    cdef DTYPE_t L1=(sqrt(3)*cg*cd-sg*cd+sqrt(2)*sd)/sqrt(6)
    cdef DTYPE_t L2=(2*sg*cd+sqrt(2)*sd)/sqrt(6)
    cdef DTYPE_t L3=(-sqrt(3)*cg*cd-sg*cd+sqrt(2)*sd)/sqrt(6)
    M[0]=L1*T1*T1+L2*N1*N1+L3*P1*P1
    M[1]=L1*T2*T2+L2*N2*N2+L3*P2*P2
    M[2]=L1*T3*T3+L2*N3*N3+L3*P3*P3
    M[3]=sqrt2*(L1*T1*T2+L2*N1*N2+L3*P1*P2)
    M[4]=sqrt2*(L1*T1*T3+L2*N1*N3+L3*P1*P3)
    M[5]=sqrt2*(L1*T2*T3+L2*N2*N3+L3*P2*P3)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void cMultipleTape_MT6(DTYPE_t*M,DTYPE_t* gamma,DTYPE_t* delta,DTYPE_t* kappa,DTYPE_t* h,DTYPE_t* sigma,Py_ssize_t n) nogil:
    for i from 0<=i<n:
        cTape_MT6(&M[i*6],gamma[i],delta[i],kappa[i],h[i],sigma[i])
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def Tape_MT6(DTYPE_t[::1] gamma,DTYPE_t[::1] delta,DTYPE_t[::1] kappa,DTYPE_t[::1] h,DTYPE_t[::1] sigma):
    cdef Py_ssize_t n=gamma.shape[0]
    if not (gamma.shape[0]==delta.shape[0] and gamma.shape[0]==kappa.shape[0] and gamma.shape[0]==h.shape[0] and gamma.shape[0]==sigma.shape[0]):
        raise ValueError('Arguments different size')
    cdef DTYPE_t[:,::1] MTs=np.empty((n,6))
    cMultipleTape_MT6(&MTs[0,0],&gamma[0],&delta[0],&kappa[0],&h[0],&sigma[0],n)
    return np.ascontiguousarray(np.asarray(MTs).T)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef MT_convert(DTYPE_t[:] MT,DTYPE_t [::1] results):
    T,N,P,E=cMT6_TNPE(MT)    
    return cTPE_convert(T,P,E,results)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def MT_output_convert(DTYPE_t[:,:] MTs):
    cdef Py_ssize_t imax=MTs.shape[1]
    cdef Py_ssize_t i
    cdef DTYPE_t [::1] g=np.empty((imax,))
    cdef DTYPE_t [::1] d=np.empty((imax,))
    cdef DTYPE_t [::1] k=np.empty((imax,))
    cdef DTYPE_t [::1] h=np.empty((imax,))
    cdef DTYPE_t [::1] s=np.empty((imax,))
    cdef DTYPE_t [::1] u=np.empty((imax,))
    cdef DTYPE_t [::1] v=np.empty((imax,))
    cdef DTYPE_t [::1] s1=np.empty((imax,))
    cdef DTYPE_t [::1] d1=np.empty((imax,))
    cdef DTYPE_t [::1] r1=np.empty((imax,))
    cdef DTYPE_t [::1] s2=np.empty((imax,))
    cdef DTYPE_t [::1] d2=np.empty((imax,))
    cdef DTYPE_t [::1] r2=np.empty((imax,))
    cdef DTYPE_t [::1] results=np.empty((13,))
    for i in range(imax):
        results=MT_convert(MTs[:,i],np.empty((13,)))
        g[i]=results[0]
        d[i]=results[1]
        k[i]=results[2]
        h[i]=results[3]
        s[i]=results[4]
        u[i]=results[5]
        v[i]=results[6]
        s1[i]=results[7]
        d1[i]=results[8]
        r1[i]=results[9]
        s2[i]=results[10]
        d2[i]=results[11]
        r2[i]=results[12]
    return {'g':np.asarray(g),'d':np.asarray(d),'k':np.asarray(k),'h':np.asarray(h),'s':np.asarray(s),'u':np.asarray(u),'v':np.asarray(v),'S1':np.asarray(s1),'D1':np.asarray(d1),'R1':np.asarray(r1),'S2':np.asarray(s2),'D2':np.asarray(d2),'R2':np.asarray(r2)}   
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef cMT6_TNPE(DTYPE_t[:] MT):
    cdef DTYPE_t [:,::1] MT33=np.empty((3,3))
    cdef DTYPE_t [::1] E=np.empty((3,))
    cdef DTYPE_t [::1] e=np.empty((3,))
    cdef DTYPE_t [::1] T=np.empty((3,))
    cdef DTYPE_t [::1] P=np.empty((3,))
    cdef DTYPE_t [::1] N=np.empty((3,))
    cdef DTYPE_t [:,:] L=np.empty((3,3))
    MT33[0,0]=MT[0]
    MT33[1,1]=MT[1]
    MT33[2,2]=MT[2]
    MT33[0,1]=MT[3]/sqrt2
    MT33[0,2]=MT[4]/sqrt2
    MT33[1,2]=MT[5]/sqrt2
    MT33[1,0]=MT[3]/sqrt2
    MT33[2,0]=MT[4]/sqrt2
    MT33[2,1]=MT[5]/sqrt2
    if check_finite>0:
        e,L=eigh(MT33,overwrite_a=True,check_finite=False)
    else:        
        e,L=eigh(MT33,overwrite_a=True)
    maxi=0
    mini=0
    cdef Py_ssize_t i
    for i in [0,1,2]:
        if e[i]>e[maxi]:
            maxi=i
        if e[i]<e[mini]:
            mini=i
    E[0]=e[maxi]
    E[1]=e[3-maxi-mini]
    E[2]=e[mini]
    T[0]=L[0,maxi]
    T[1]=L[1,maxi]
    T[2]=L[2,maxi]
    N[0]=L[0,3-maxi-mini]
    N[1]=L[1,3-maxi-mini]
    N[2]=L[2,3-maxi-mini]
    P[0]=L[0,mini]
    P[1]=L[1,mini]
    P[2]=L[2,mini]
    return T,N,P,E
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef MT6_TNPE(DTYPE_t[:,:] MTs):
    cdef Py_ssize_t imax=MTs.shape[1]
    cdef Py_ssize_t i
    cdef DTYPE_t [:,::1] T=np.empty((3,imax))
    cdef DTYPE_t [:,::1] N=np.empty((3,imax))
    cdef DTYPE_t [:,::1] P=np.empty((3,imax))
    cdef DTYPE_t [:,::1] E=np.empty((3,imax))
    for i in range(imax):
        t,n,p,e=cMT6_TNPE(MTs[:,i])
        T[0,i]=t[0]
        N[0,i]=n[0]
        P[0,i]=p[0]
        E[0,i]=e[0]
        T[1,i]=t[1]
        N[1,i]=n[1]
        P[1,i]=p[1]
        E[1,i]=e[1]
        T[2,i]=t[2]
        N[2,i]=n[2]
        P[2,i]=p[2]
        E[2,i]=e[2]
    return np.asarray(T),np.asarray(N),np.asarray(P),np.asarray(E)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef E_GD(DTYPE_t[:,:]E):
    cdef Py_ssize_t imax=E.shape[1]
    cdef Py_ssize_t i
    cdef DTYPE_t [::1] g=np.empty((imax,))
    cdef DTYPE_t [::1] d=np.empty((imax,))
    for i in range(imax):
        cE_gd(E[:,i],&g[i],&d[i])
    return np.asarray(g),np.asarray(d)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef cSingleTP_SDR(DTYPE_t [:] T,DTYPE_t[:]P):
    cdef DTYPE_t t0 = T[0]
    cdef DTYPE_t t1 = T[1]
    cdef DTYPE_t t2 = T[2]
    cdef DTYPE_t p0 = P[0]
    cdef DTYPE_t p1 = P[1]
    cdef DTYPE_t p2 = P[2]
    cdef DTYPE_t Nt = sqrt((t0+p0)*(t0+p0)+(t1+p1)*(t1+p1)+(t2+p2)*(t2+p2))
    cdef DTYPE_t St = sqrt((t0-p0)*(t0-p0)+(t1-p1)*(t1-p1)+(t2-p2)*(t2-p2))
    cdef DTYPE_t N0 = (t0+p0)/Nt
    cdef DTYPE_t N1 = (t1+p1)/Nt
    cdef DTYPE_t N2 = (t2+p2)/Nt
    cdef DTYPE_t S0 = (t0-p0)/St
    cdef DTYPE_t S1 = (t1-p1)/St
    cdef DTYPE_t S2 = (t2-p2)/St
    cdef DTYPE_t strike1
    cdef DTYPE_t dip1
    cdef DTYPE_t rake1
    cN_SDR(N0, N1, N2, S0, S1, S2, &strike1, &dip1, &rake1)
    return strike1, dip1, rake1

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef TP_SDR(DTYPE_t[:,:]T,DTYPE_t[:,:]P):
    cdef Py_ssize_t imax=T.shape[1] 
    cdef Py_ssize_t i
    cdef DTYPE_t [::1] s=np.empty((imax,))
    cdef DTYPE_t [::1] d=np.empty((imax,))
    cdef DTYPE_t [::1] r=np.empty((imax,))
    for i in range(imax):
        s[i],d[i],r[i]=cSingleTP_SDR(T[:,i],P[:,i])
    return np.asarray(s),np.asarray(d),np.asarray(r)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void csingleSDR_SDR(DTYPE_t s,DTYPE_t d,DTYPE_t r,DTYPE_t*s2,DTYPE_t*d2,DTYPE_t*r2):
    cdef DTYPE_t ck=cos(s)
    cdef DTYPE_t cs=cos(r)
    cdef DTYPE_t sk=sin(s)
    cdef DTYPE_t ss=sin(r)
    cdef DTYPE_t h=cos(d)
    cdef DTYPE_t sh=sqrt(1-h*h)
    cdef DTYPE_t NT=sqrt((ck*cs+sk*h*ss-sk*sh)*(ck*cs+sk*h*ss-sk*sh)+(sk*cs-ck*h*ss+ck*sh)*(sk*cs-ck*h*ss+ck*sh)+(-sh*ss-h)*(-sh*ss-h))
    cdef DTYPE_t NP=sqrt((ck*cs+sk*h*ss+sk*sh)*(ck*cs+sk*h*ss+sk*sh)+(sk*cs-ck*h*ss-ck*sh)*(sk*cs-ck*h*ss-ck*sh)+(-sh*ss+h)*(-sh*ss+h))
    cdef DTYPE_t T1=(ck*cs+sk*h*ss-sk*sh)/NT
    cdef DTYPE_t T2=(sk*cs-ck*h*ss+ck*sh)/NT
    cdef DTYPE_t T3=(-sh*ss-h)/NT
    cdef DTYPE_t P1=(ck*cs+sk*h*ss+sk*sh)/NP
    cdef DTYPE_t P2=(sk*cs-ck*h*ss-ck*sh)/NP
    cdef DTYPE_t P3=(-sh*ss+h)/NP
    cdef DTYPE_t N0=(T1-P1)
    cdef DTYPE_t N1=(T2-P2)
    cdef DTYPE_t N2=(T3-P3)
    cdef DTYPE_t S0=(T1+P1)
    cdef DTYPE_t S1=(T2+P2)
    cdef DTYPE_t S2=(T3+P3)
    cdef DTYPE_t St=sqrt(S0*S0+S1*S1+S2*S2)
    cdef DTYPE_t Nt=sqrt(N0*N0+N1*N1+N2*N2)
    cN_SDR(S0/St,S1/St,S2/St,N0/Nt,N1/Nt,N2/Nt,s2,d2,r2)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef SDR_SDR(DTYPE_t[::1] s,DTYPE_t[::1] d,DTYPE_t[::1] r):
    cdef Py_ssize_t imax=s.shape[0]
    cdef Py_ssize_t i
    cdef DTYPE_t [::1] s2=np.empty((imax,))
    cdef DTYPE_t [::1] d2=np.empty((imax,))
    cdef DTYPE_t [::1] r2=np.empty((imax,))
    for i in range(imax):
        csingleSDR_SDR(s[i],d[i],r[i],&s2[i],&d2[i],&r2[i])
    return np.asarray(s2),np.asarray(d2),np.asarray(r2)

#Bi-axes
cpdef list isotropic_c(DTYPE_t l=1.,DTYPE_t mu=1.,list c=[]):
    if len(c)==21: #Calculate isotropic approacximation
            #Eqns 81a and 81b from chapman and leaney 2011
            mu=((c[0]+c[6]+c[11])+3*(c[15]+c[18]+c[20])-(c[1]+c[2]+c[7]))/15
            l=((c[0]+c[6]+c[11])-2*(c[15]+c[18]+c[20])+4*(c[1]+c[2]+c[7]))/15
    n=l+2*mu
    return [n,l,l,0,0,0,n,l,0,0,0,n,0,0,0,mu,0,0,mu,0,mu]

def MT6_biaxes(DTYPE_t[:]MT6,list c):
    lambda2mu=(3*(c[0]+c[6]+c[11])+4*(c[15]+c[18]+c[20])+2*(c[1]+c[2]+c[7]))/15
    mu=((c[0]+c[6]+c[11])+3*(c[15]+c[18]+c[20])-(c[1]+c[2]+c[7]))/15
    l=lambda2mu-2*mu
    T,N,P,E=cMT6_TNPE(MT6)
    isotropic=(l+mu)*E[1]/mu-l*(E[0]+E[2])/(2*mu)
    if is_isotropic_c(c):
        explosion=isotropic
    else:
        def isotropic_solve(iso):
            iso6=np.squeeze(iso)*np.array([[1],[1],[1],[0],[0],[0]])
            if iso6.shape!=MT6.shape:
                iso6=iso6.T
            T,N,P,E=cMT6_TNPE(MT6c_D6(np.squeeze(MT6-iso6),c).flatten())
            return E[1]
        explosion=fsolve(isotropic_solve,isotropic)

    explosion6=np.squeeze(explosion)*np.array([[1],[1],[1],[0],[0],[0]])
    if explosion6.shape!=MT6.shape:
        explosion6=explosion6.T
    T,N,P,E=cMT6_TNPE(MT6c_D6(np.squeeze(MT6-explosion6),c).flatten())   
    area_displacement = E[0]-E[2]
    phi=np.zeros((3,2))  
    if area_displacement!=0:      # to avoid undefined
        cphi=np.squeeze(np.sqrt(E[0]/area_displacement))
        sphi=np.squeeze(np.sqrt(-E[2]/area_displacement))
        phi[:,0]=np.array(cphi*T+sphi*P).flatten()
        phi[:,1]=np.array(cphi*T-sphi*P).flatten()
    return phi,explosion,area_displacement

cpdef MT6c_D6(mt6,list c=isotropic_c(l=1,mu=1)):
    mtvoigt=mt6[np.array([0,1,2,5,4,3])]
    mtvoigt=np.matrix(mtvoigt)
    if mtvoigt.shape[1]==6:
        mtvoigt=mtvoigt.T
    #Convert to voigt
    dvoigt=np.linalg.solve(np.matrix(c21_cvoigt(c)),mtvoigt)
    return np.asarray(dvoigt[np.array([0,1,2,5,4,3])])

cpdef bool is_isotropic_c(list c):
    cdef DTYPE_t tol = 1.e-6*c_norm(c);
    return ((fabs(c[ 3])<tol)and(fabs(c[ 4])<tol)and(fabs(c[ 5])<tol)and \
            (fabs(c[ 8])<tol)and(fabs(c[ 9])<tol)and(fabs(c[10])<tol)and \
            (fabs(c[12])<tol)and(fabs(c[13])<tol)and(fabs(c[14])<tol)and \
            (fabs(c[16])<tol)and(fabs(c[17])<tol)and(fabs(c[19])<tol)and \
            (fabs(c[ 0]-c[ 6])<tol)and(fabs(c[ 6]-c[11])<tol)and \
            (fabs(c[15]-c[18])<tol)and(fabs(c[18]-c[20])<tol)and \
            (fabs(c[ 1]-c[ 2])<tol)and(fabs(c[ 2]-c[ 7])<tol)and \
            (fabs(c[0]-c[1]-2*c[15])<tol))

cpdef DTYPE_t [:,::1] c21_cvoigt(list c):
    return np.array([[c[0],c[1],c[2],sqrt2*c[3],sqrt2*c[4],sqrt2*c[5]],
                     [c[1],c[6],c[7],sqrt2*c[8],sqrt2*c[9],sqrt2*c[10]],
                     [c[2],c[7],c[11],sqrt2*c[12],sqrt2*c[13],sqrt2*c[14]],
                     [sqrt2*c[3],sqrt2*c[8],sqrt2*c[12],2*c[15],2*c[16],2*c[17]],
                     [sqrt2*c[4],sqrt2*c[9],sqrt2*c[13],2*c[16],2*c[18],2*c[19]],
                     [sqrt2*c[5],sqrt2*c[10],sqrt2*c[14],2*c[17],2*c[19],2*c[20]]])

cpdef DTYPE_t c_norm(list c):
    return sqrt(c[0]**2+c[6]**2+c[11]**2+2*(c[1]**2+c[2]**2+c[7]**2)+
                 4*(c[3]**2+c[4]**2+c[5]**2+c[8]**2+c[9]**2+c[10]**2+
                    c[12]**2+c[13]**2+c[14]**2+c[15]**2+c[18]**2+c[20]**2)+
                 8*(c[16]**2+c[17]**2+c[19]**2));  

# Test functions - Not Documented

class cMomentTensorConvertTestCase(TestCase):
    def test_cTape_MT6(self):
        cdef DTYPE_t[::1] m=np.empty((6))
        cTape_MT6(&m[0],0.12,0.43,0.76,0.63,0.75)
        self.assertAlmostEqual(m[0],-0.3637,4)
        self.assertAlmostEqual(m[1],0.4209,4)
        self.assertAlmostEqual(m[2],0.6649,4)
        self.assertAlmostEqual(m[3],0.3533,4)
        self.assertAlmostEqual(m[4],-0.1952,4)
        self.assertAlmostEqual(m[5],-0.2924,4)
    def test_c_cE_tk(self):
        results=np.array([0.,0.,0.,0.,0.,0.,0.])
        E=np.array([1.,0.,-1.])
        results=cE_tk(E,results)
        self.assertAlmostEqual(results[5],0)
        self.assertAlmostEqual(results[6],0)
        E=np.array([1.,1.,1.])
        results=cE_tk(E,results)
        self.assertAlmostEqual(results[5],1.)
        self.assertAlmostEqual(results[6],0.)
        E=np.array([1.,-1.,-1.])
        results=cE_tk(E,results)
        self.assertAlmostEqual(results[5],-0.2)#y
        self.assertAlmostEqual(results[6],-0.8)#x
    def test_c_ctk_uv(self):
        results=np.array([0.,0.,0.,0.,0.,0.,0.])
        E=np.array([1.,0.,-1.])
        results=cE_tk(E,results)
        results=ctk_uv(results)
        self.assertAlmostEqual(results[5],0)
        self.assertAlmostEqual(results[6],0)
        E=np.array([1.,1.,1.])
        results=cE_tk(E,results)
        results=ctk_uv(results)
        self.assertAlmostEqual(results[5],0)
        self.assertAlmostEqual(results[6],1)
        E=np.array([1.,-1.,-1.])
        results=cE_tk(E,results)
        results=ctk_uv(results)
        self.assertAlmostEqual(results[5],-4./3)
        self.assertAlmostEqual(results[6],-1./3)
    def test_c_cE_gd(self):
        E=np.array([1.,0.,-1.])
        cdef DTYPE_t g
        cdef DTYPE_t d
        cE_gd(E,&g,&d)
        self.assertAlmostEqual(g,0)
        self.assertAlmostEqual(d,0)
        E=np.array([1.,1.,1.])
        cE_gd(E,&g,&d)
        self.assertAlmostEqual(g,0)
        self.assertAlmostEqual(d,pi/2)
        E=np.array([1.,-1.,-1.])
        cE_gd(E,&g,&d)
        self.assertAlmostEqual(g,-0.523598775598299)
        self.assertAlmostEqual(d,-0.339836909454122)
    def test_c_cN_sdr(self):
        E=np.array([1.,0.,-1.])
        cdef DTYPE_t s
        cdef DTYPE_t d
        cdef DTYPE_t r
        cN_SDR(0.,0.5/np.sqrt(1.25),1.0/np.sqrt(1.25),0.5/np.sqrt(1.25),0,1.0/np.sqrt(1.25),&s,&d,&r)
        self.assertAlmostEqual(s,pi)
        self.assertAlmostEqual(d,0.463647609000806)
        self.assertAlmostEqual(r,1.35080834939944)
        cN_SDR(0.5/np.sqrt(1.25),0,1.0/np.sqrt(1.25),0.,0.5/np.sqrt(1.25),1.0/np.sqrt(1.25),&s,&d,&r)
        self.assertAlmostEqual(s,pi/2)
        self.assertAlmostEqual(d,0.463647609000806)
        self.assertAlmostEqual(r,1.79078430419036)
        N0 = 0.70710678
        N1 = 0.11957316
        N2 = -0.69692343
        S0 = 0.70710678
        S1 = -0.11957316
        S2 = 0.69692343
        cN_SDR(N0, N1, N2, S0, S1, S2, &s, &d, &r)
        T = np.array([1., 0., 0.])
        P = np.array([0.,
                      0.16910198,
                      -0.98559856])
        # Test single TP - N conversion
        s2, d2, r2 = cSingleTP_SDR(T, P)
        self.assertEqual(s, s2)
        self.assertEqual(d, d2)
        self.assertEqual(r, r2)
        try:
            self.assertAlmostEqual(s, 279.5980303*pi/180)
            self.assertAlmostEqual(d, 45.81931182*pi/180)
            self.assertAlmostEqual(r, -76.36129238206001*pi/180)
        except Exception:
            self.assertAlmostEqual(s, 80.4019697*pi/180)
            self.assertAlmostEqual(d, 45.81931182*pi/180)
            self.assertAlmostEqual(r,-103.63870728*pi/180)

    def test_c_cTP_SDR(self):
        T=np.array([0.235702260395516,
         0.235702260395516,
         0.942809041582063])
        P=np.array([0.707106781186547,
        -0.707106781186547,
                         0])
        results=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.])
        results=cTP_SDR(T,P,results)
        self.assertAlmostEqual(results[2],3.6052402625906)
        self.assertAlmostEqual(results[3],0.666666666666667)
        self.assertAlmostEqual(results[4],1.10714871779409)
        try:
            self.assertAlmostEqual(results[7],206.565051177078)
            self.assertAlmostEqual(results[8],48.1896851042214)
            self.assertAlmostEqual(results[9],63.434948822922)
            self.assertAlmostEqual(results[10],63.434948822922)
            self.assertAlmostEqual(results[11],48.1896851042214)
            self.assertAlmostEqual(results[12],116.565051177078)
        except Exception:
            self.assertAlmostEqual(results[10],206.565051177078)
            self.assertAlmostEqual(results[11],48.1896851042214)
            self.assertAlmostEqual(results[12],63.434948822922)
            self.assertAlmostEqual(results[7],63.434948822922)
            self.assertAlmostEqual(results[8],48.1896851042214)
            self.assertAlmostEqual(results[9],116.565051177078)

    def test_c_cSingleTP_SDR(self):
        T = np.array([0.235702260395516,
                      0.235702260395516,
                      0.942809041582063])
        P = np.array([0.707106781186547,
                      -0.707106781186547,
                      0])
        s, d, r = cSingleTP_SDR(T, P)
        try:
            self.assertAlmostEqual(s, 3.6052402625906)
            self.assertAlmostEqual(d, 0.666666666666667)
            self.assertAlmostEqual(r, 1.10714871779409)
        except Exception:
            self.assertAlmostEqual(s, 63.434948822922*pi/180)
            self.assertAlmostEqual(d, 48.1896851042214*pi/180)
            self.assertAlmostEqual(r, 116.565051177078*pi/180)
        T = np.array([1., 0., 0.])
        P = np.array([0.,
                      0.16910198,
                      -0.98559856])
        s, d, r = cSingleTP_SDR(T, P)
        try:
            self.assertAlmostEqual(s, 279.5980303*pi/180)
            self.assertAlmostEqual(d, 45.81931182*pi/180)
            self.assertAlmostEqual(r, -76.36129238206001*pi/180)
        except Exception:
            self.assertAlmostEqual(s, 80.4019697*pi/180)
            self.assertAlmostEqual(d, 45.81931182*pi/180)
            self.assertAlmostEqual(r,-103.63870728*pi/180)

    def test_c_MT6_TNPE(self):
        MT=np.array([[1.,0.,-1.,0.,0.,0.],[0,2.0,-1.0,0.,1.0,0.]]).T
        T,N,P,E=MT6_TNPE(MT)
        self.assertAlmostEqual(E[0,0],1)
        self.assertAlmostEqual(E[1,0],0)
        self.assertAlmostEqual(E[2,0],-1)
        self.assertAlmostEqual(np.abs(T[0,0]),1)
        self.assertAlmostEqual(np.abs(T[1,0]),0)
        self.assertAlmostEqual(np.abs(T[2,0]),0)
        self.assertAlmostEqual(np.abs(N[0,0]),0)
        self.assertAlmostEqual(np.abs(N[1,0]),1)
        self.assertAlmostEqual(np.abs(N[2,0]),0)
        self.assertAlmostEqual(np.abs(P[0,0]),0)
        self.assertAlmostEqual(np.abs(P[1,0]),0)
        self.assertAlmostEqual(np.abs(P[2,0]),1)
        #Second Event
        self.assertAlmostEqual(E[0,1],2)
        self.assertAlmostEqual(E[1,1],0.366025403784439 )
        self.assertAlmostEqual(E[2,1],-1.36602540378444)
        self.assertAlmostEqual(np.abs(T[0,1]),0)
        self.assertAlmostEqual(np.abs(T[1,1]),1)
        self.assertAlmostEqual(np.abs(T[2,1]),0)
        self.assertAlmostEqual(np.abs(N[0,1]),0.888073833977115)
        self.assertAlmostEqual(np.abs(N[1,1]),0)
        self.assertAlmostEqual(np.abs(N[2,1]),0.459700843380983)
        self.assertAlmostEqual(np.abs(P[0,1]),0.459700843380983)
        self.assertAlmostEqual(np.abs(P[1,1]),0)
        self.assertAlmostEqual(np.abs(P[2,1]),0.888073833977115)
    def test_c_E_GD(self):
        E=np.array([[1.,0.,-1.],[1.,1.,1.],[1.,-1.,-1.],[-1.,-1.,-1.]]).T
        g,d=E_GD(E)
        self.assertAlmostEqual(g[0],np.array([0.]))
        self.assertAlmostEqual(d[0],np.array([0.]))
        self.assertAlmostEqual(g[1],np.array([0.]))
        self.assertAlmostEqual(d[1],np.array([pi/2]))
        self.assertAlmostEqual(g[2],np.array([-0.523598775598299]))
        self.assertAlmostEqual(d[2],np.array([-0.339836909454122]))
        self.assertAlmostEqual(g[3],np.array([0.]))
        self.assertAlmostEqual(d[3],np.array([-pi/2]))
    def test_c_TP_SDR(self):
        T=np.array([[0.235702260395516,0.],[0.235702260395516,1.0],[0.942809041582063,0.]])
        P=np.array([[0.707106781186547,1.],[-0.707106781186547,0],[0.,0.]])
        s,d,r=TP_SDR(T,P)
        self.assertAlmostEqual(s[0],np.array([1.1071487177940911]))
        self.assertAlmostEqual(d[0],np.array([0.84106867056793]))
        self.assertAlmostEqual(r[0],np.array([2.0344439357957032]))
        self.assertAlmostEqual(s[1],np.array([5.497787143782138]))
        self.assertAlmostEqual(d[1],np.array([pi/2]))
        self.assertAlmostEqual(r[1],np.array([-pi]))
    def test_c_SDR_SDR(self):
        s2,d2,r2=SDR_SDR(np.array([206.565051177078*pi/180]),np.array([48.1896851042214*pi/180]),np.array([63.434948822922*pi/180]))
        self.assertAlmostEqual(s2,63.434948822922*pi/180)
        self.assertAlmostEqual(d2,48.1896851042214*pi/180)
        self.assertAlmostEqual(r2,116.565051177078*pi/180)
        s2,d2,r2=SDR_SDR(np.array([63.434948822922*pi/180]),np.array([48.1896851042214*pi/180]),np.array([116.565051177078*pi/180]))
        self.assertAlmostEqual(s2,206.565051177078*pi/180)
        self.assertAlmostEqual(d2,48.1896851042214*pi/180)
        self.assertAlmostEqual(r2,63.434948822922*pi/180)

def test_suite(verbosity=2):
    suite = [unittest.TestLoader().loadTestsFromTestCase(cMomentTensorConvertTestCase), ]
    suite = unittest.TestSuite(suite)
    return suite